#include "v850es.h"
#include "integer.h"
#include "taskmgr.h"
#include "tff.h"
#include "glcd.h"


#define	PLCD_VRAM	0x100000	/* Graphic buffer */
#define	PLCD_REG	0x11FFE0	/* S1D13705 control register */

#define PKANJI12	0x15000		/* Kanji font base address (internal ROM) */
#define	PANK12		0x3E000		/* ANK font base address (internal ROM) */

static
BYTE IoBuff[2048];







BOOL gd_boxfill (
	const RECT* rect,
	BYTE idx
)
{
	BYTE *p;
	UINT n, w, d, l, t, r, b;


	l = rect->l; t = rect->t; r = rect->r; b = rect->b;
	if (t > b || l > r || b >= LCD_HEIGHT || r >= LCD_WIDTH) return FALSE;
	d = (UINT)(idx << 8) | idx;
	p = (BYTE*)(PLCD_VRAM + t * LCD_WIDTH + l);
	for ( ; t <= b; t++) {
		w = r - l + 1;
		if (l & 1) {
			*p++ = (BYTE)d; w--;
		}
		for (n = w / 2; n; n--) { *(WORD*)p = (WORD)d; p += 2; }
		if (w & 1) *p++ = (BYTE)d;
		p += LCD_WIDTH - r + l - 1;
	}
	return TRUE;
}




BYTE gd_get_pixel (
	POINT pnt
) {
	return (pnt.x < LCD_WIDTH && pnt.y < LCD_HEIGHT) ? *(BYTE*)(PLCD_VRAM + pnt.y * LCD_WIDTH + pnt.x) : 0xFF;
}




void gd_put_pixel (
	POINT pnt,
	BYTE index
) {
	if (pnt.x < LCD_WIDTH && pnt.y < LCD_HEIGHT)
		*(BYTE*)(PLCD_VRAM + pnt.y * LCD_WIDTH + pnt.x) = index;
}




BOOL gd_put_kanji (
	POINT pnt,
	WORD sjis,
	BYTE idx
) {
	BYTE *p, *f, *ft;
	UINT h, v, t, s, e;


	if (pnt.x + KANJI_W > LCD_WIDTH || pnt.y + KANJI_H > LCD_HEIGHT) return FALSE;
	p = (BYTE*)(PLCD_VRAM + pnt.y * LCD_WIDTH + pnt.x);
	f = (BYTE*)(PKANJI12 + 17);
	h = *f++;
	ft = f + h * 4;
	t = 0;
	for (; h; h--) {
		s = LD_WORD(f); e = LD_WORD(f+2); f += 4;
		if (sjis >= s && sjis <= e) break;
		t += e - s + 1;
	}
	if (h) t += sjis - s;
	t *= KANJI_H * 2;
	f = ft + t;
	for (v = KANJI_H; v; v--) {
		t = *f++; t = (t << 8) + *f++;
		for (h = 0x8000; h > (0x8000 >> KANJI_W); h >>= 1) {
			if (t & h) *p = idx;
			p++;
		}
		p += LCD_WIDTH - KANJI_W;
	}

	return TRUE;
}




BOOL gd_put_ank (
	POINT pnt,
	BYTE ank,
	BYTE idx
) {
	BYTE *p, *f;
	UINT v, h, t;


	if (pnt.x + ANK_W > LCD_WIDTH || pnt.y + ANK_H > LCD_HEIGHT) return FALSE;
	p = (BYTE*)(PLCD_VRAM + pnt.y * LCD_WIDTH + pnt.x);
	f = (BYTE*)(PANK12 + 17 + ank * ANK_H);

	for (v = ANK_H; v; v--) {
		t = *f++;
		for (h = 0x80; h > (0x80 >> ANK_W); h >>= 1) {
			if (t & h) *p = idx;
			p++;
		}
		p += LCD_WIDTH - ANK_W;
	}

	return TRUE;
}




BOOL gd_put_7seg (
	POINT pnt,
	BYTE fnt,
	BYTE idx
) {
	static const WORD font[] = {
/*0*/	0x7FC0,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x7FC0,
/*1*/	0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,
/*2*/	0x7FC0,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x7FC0,0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,0x7FC0,
/*3*/	0x7FC0,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x7FC0,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x7FC0,
/*4*/	0x0020,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x7FE0,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,
/*5*/	0x7FC0,0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,0x7FC0,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x7FC0,
/*6*/	0x7FC0,0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,0xFFC0,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x7FC0,
/*7*/	0x7FC0,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,
/*8*/	0x7FC0,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x7FC0,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x7FC0,
/*9*/	0x7FC0,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x8020,0x7FE0,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x7FC0,
/*:*/	0x0000,0x0000,0x0000,0x0000,0xC000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0xC000,0x0000,0x0000,0x0000,0x0000
	};
	BYTE *p;
	const WORD *f;
	UINT v, h, t;


	if (pnt.x + SEG7_W > LCD_WIDTH || pnt.y + SEG7_H > LCD_HEIGHT) return FALSE;
	if (fnt >= 11) return TRUE;
	p = (BYTE*)(PLCD_VRAM + pnt.y * LCD_WIDTH + pnt.x);
	f = &font[fnt * SEG7_H];

	for (v = SEG7_H; v; v--) {
		t = *f++;
		for (h = 0x8000; h > (0x8000 >> SEG7_W); h >>= 1) {
			if (t & h) *p = idx;
			p++;
		}
		p += LCD_WIDTH - SEG7_W;
	}

	return TRUE;
}




void gd_put_str (
	POINT pnt,
	BYTE* str,
	UINT width,
	BYTE idx
) {
	UINT c;


	for (;;) {
		c = *str++;
		if (c == 0) break;
		if ((c >= 0x81 && c <= 0x9F) || (c >= 0xE0 && c <= 0xFC)) {
			if (*str == 0) break;
			c = (c << 8) + *str++;
			if (width < KANJI_W || !gd_put_kanji(pnt, (WORD)c, idx)) break;
			pnt.x += KANJI_W;
			width -= KANJI_W;
		} else {
			if (width < ANK_W || !gd_put_ank(pnt, (BYTE)c, idx)) break;
			pnt.x += ANK_W;
			width -= ANK_W;
		}
	}
}




BOOL gd_load_bmpfile (
	FIL* fil,
	POINT pnt
) {
	volatile BYTE *vp;
	BYTE *lp, *bp;
	UINT rb;
	DWORD iw, iwf, ih, ly, d, n, m;
	FRESULT res;


	if (pnt.x >= LCD_WIDTH || pnt.y >= LCD_HEIGHT) return FALSE;

	/* Load File Header */
	lp = IoBuff;
	res = f_read(fil, lp, 14, &rb);
	if (res || rb != 14 || LD_WORD(lp) != 0x4D42) return FALSE;

	/* Load Information Header */
	res = f_read(fil, lp, 40, &rb);
	if (res || rb != 40 || LD_WORD(lp+14) != 8 || LD_DWORD(lp+16) != 0) return FALSE;
	iw = LD_DWORD(lp+4);
	ih = LD_DWORD(lp+8);
	if (iw > sizeof(IoBuff)) return FALSE;

	/* Load Pallete */
	res = f_read(fil, lp, 4 * 256, &rb);
	if (res || rb != 4 * 256) return FALSE;
	vp = (BYTE*)PLCD_REG;
	vp[0x15] = 0;
	for (n = 0; n < 240; n++) {
		d = LD_DWORD(lp); lp += 4;
		vp[0x17] = (BYTE)(d >> 16);
		vp[0x17] = (BYTE)(d >> 8);
		vp[0x17] = (BYTE)d;
	}

	/* Load bitmap */
	iwf = (iw + 3) & ~3;
	ly = pnt.y + ih;
	do {
		n = (ih >= sizeof(IoBuff) / iwf) ? sizeof(IoBuff) / iwf : ih;
		ih -= n;
		lp = IoBuff;
		res = f_read(fil, lp, iwf * n, &rb);
		if (res || rb != iwf * n) return FALSE;
		do {
			if (--ly < LCD_HEIGHT) {
				vp = (BYTE*)(PLCD_VRAM + ly * LCD_WIDTH + pnt.x);
				bp = lp;
				m = iw;
				if (pnt.x + m > LCD_WIDTH) m = LCD_WIDTH - pnt.x;
				if (pnt.x & 1) { *vp++ = *bp++; m--; }
				for (d = m / 2; d; d--) {
					*(WORD*)vp = *(WORD*)bp;
					vp += 2; bp += 2;
				}
				if (m & 1) *vp = *bp;
			}
			lp += iwf;
		} while (--n);
	} while (ih);

	return TRUE;
}




void gd_enable (void)
{
	UINT n;
	static BOOL init;
	static const BYTE defpal[] = {
		0,0,0, 167,0,0, 0,167,0, 167,167,0,
		0,0,167, 167,0,167, 0,167,167, 167,167,167,
		87,87,87, 255,0,0, 0,255,0, 255,255,0,
		0,0,255, 255,0,255, 0,255,255, 255,255,255 };
	WORD *wp;
	BYTE *reg = (BYTE*)PLCD_REG;



	if (!init) {
		sleep(20);
		reg[3] = 0x08;			/* Enter power save mode */
		reg[1] = 0x23;			/* LCD data format = STN/Single/Color */
		reg[2] = 0xC0;			/* Bit/pix = 8, pclk = clki */
		reg[4] = LCD_WIDTH / 8 - 1;				/* H disp = 320 */
		reg[5] = LCD_HEIGHT - 1; reg[6] = 0;	/* V disp = 240 */
		reg[8] = 152/8-4;		/* H blank = 152px */
		reg[10] = 0;			/* V blank = 0px */
		reg[11] = 0;			/* MOD toggle interval = 1 */
		reg[12] = 0; reg[13] = 0;	/* Display start address = 0 */
		reg[17] = 0;			/* Disable virtual screen */
		reg[18] = LCD_HEIGHT; reg[19] = 0;		/* Plane 1 height */
		reg[27] = 0;			/* Landscape mode */
		for (wp = (WORD*)PLCD_VRAM, n = LCD_WIDTH * LCD_HEIGHT / 2; n; *wp++ = ((WORD)BLACK << 8) + BLACK, n--);	/* Clear screen */
		init = TRUE;
	}
	if (!PCM.3) {
		reg[3] = 0x0B;			/* Enable LCD signals */
		sleep(20);				/* Delay 20ms */
		reg[3] = 0x03;			/* Enable display */
		reg[0x15] = 0xF0;		/* Load system palette */
		for (n = 0; n < 16 * 3; n++) reg[0x17] = defpal[n];
		PCM.3 = 1;				/* Enable backlight */
	}
}




void gd_disable (void)
{
	BYTE *reg = (BYTE*)PLCD_REG;


	PCM.3 = 0;			/* Disable backlight */
	reg[3] = 0x00;		/* Enter power save mode */
}



BOOL gd_check (void)
{
	return PCM.3 ? TRUE : FALSE;
}



