#include <string.h>
#include <ctype.h>
#include "integer.h"
#include "tff.h"
#include "taskmgr.h"
#include "monitor.h"
#include "glcd.h"
#include "gui.h"
#include "play.h"
#include "iic.h"


#define CNF_FILE	"00config.txt"

/* Button code from tablet task */
volatile BYTE CmdButton;

/* Message to tablet task */
static UINT TabCmd;

/* Playback control */
BYTE Volume, Attenuation;
static BYTE Gain;

/* Display control message from play task */
volatile UINT GuiCmd;
volatile UINT GuiCurrTrk, GuiNumTrks;	/* Track number n of m */
volatile UINT GuiTrkPos, GuiTrkSize;	/* Track playing point n of m */
volatile char GuiFile[30], GuiTitle[64], GuiArtist[64];	/* File name, Title and Artist */
volatile BYTE GuiContrast;				/* Contrast setting for LCD display */

/* Display control message from timer task */
volatile INT PlayTime, PlayState, TmrGui;	/* Playback time in ms, Playback state and GUI timer */
volatile WORD GuiLevel[3];	/* ADC data [L-ch, R-ch, temperaure] */
volatile POINT GuiPoint;	/* Tablet pointed position */

/* Display configuration parameters */
static RECT RectLevel[2], RectProg, RectTitle, RectTime, RectTrack;
static BYTE ColLevel[4], ColProg[2], ColTitle[2], ColTime[2], ColTrack[2];
static BYTE DirLevel[2], SmallTime, TitleAlign;
static char CfgFile[30];	/* Current loaded configuration file */
static WORD LevelLv[2], LevelPk[2], LevelPkTmr[2];	/* Level meter work area */
static UINT GuiLevelGain;		/* Level meter gain setting */

/* Button strings (in play, in stop) */
static const char menustr1[] = "O!\tO!ׯ\t!ׯ\t!\t~\tĐ\t";
static const char menustr2[] = "S!߰\t!߰\t-\t-\t-\tĐ\t";
static BYTE ColButton[2] = { WHITE, BLACK };	/* Button color [button, letter] */



/*------------------------------------------------------------------*/
/* Replace file name in a full path name                            */
/*------------------------------------------------------------------*/

static
void replace_filename (
	char* dst,			/* Destinations buffer to store created full path name */
	int dstsize,		/* Size of the buffer */
	const char* dir,	/* Source full path name includes file name */
	const char* file	/* Source File name to replase */
) {
	UINT i, n;
	char c;

	n = i = 0;
	while (i < dstsize - 1) {
		dst[i++] = c = *dir++;
		if ((BYTE)c <= ' ') break;
		if (c == '/') n = i;
	}
	while (*file == ' ') file++;
	while (n < dstsize - 1 && (BYTE)*file > ' ') dst[n++] = *file++;
	dst[n] = 0;
}



/*------------------------------------------------------------------*/
/* Get four numbers from a string                                   */
/*------------------------------------------------------------------*/

static
BOOL get_rect (
	const char** str,	/* Pointer to pointer to the string */
	RECT* rect	 		/* Pointer to the RECT structure to store the numbers */
) {
	long val;


	if (!xatoi(str, &val)) return FALSE;
	rect->l = val;
	if (!xatoi(str, &val)) return FALSE;
	rect->t = val;
	if (!xatoi(str, &val)) return FALSE;
	rect->r = val;
	if (!xatoi(str, &val)) return FALSE;
	rect->b = val;
	return TRUE;
}



/*------------------------------------------------------------------*/
/* Get two numbers from a string                                    */
/*------------------------------------------------------------------*/

static
BOOL get_point (
	const char** str,	/* Pointer to pointer to the string */
	POINT* pnt			/* Pointer to the POINT structure to store the numbers */
) {
	long val;


	if (!xatoi(str, &val)) return FALSE;
	pnt->x = val;
	if (!xatoi(str, &val)) return FALSE;
	pnt->y = val;
	return TRUE;
}


/*------------------------------------------------------------------*/
/* Get specified number of numbers from a string                    */
/*------------------------------------------------------------------*/

static
BOOL get_color (
	const char** str,	/* Pointer to pointer to the string */
	BYTE* dst,			/* Pointer to the BYTE array to store the numbers */
	UINT num			/* Number of numbers to get */
) {
	long val;


	while (num--) {
		if (!xatoi(str, &val)) return FALSE;
		*dst++ = (BYTE)val;
	}
	return TRUE;
}



/*------------------------------------------------------------------*/
/* Load configuration parameters from a file                        */
/*------------------------------------------------------------------*/

static
void load_cfgfile (
	const char* cfgfile		/* Configuration file */
) {
	FIL fil;
	char c, line[50], imgfile[30];
	const char *pc, *ps;
	POINT imgofs;
	BYTE lvl;
	UINT n, i;
	WORD rb;
	const WORD *w;
	long val;
	RECT rect;
	const char *kwd =
		"ImageFile:\0"		/* 0 */
		"RectLevelL:\0"		/* 1 */
		"RectLevelR:\0"		/* 2 */
		"RectTitle:\0"		/* 3 */
		"RectProgress:\0"	/* 4 */
		"RectTime:\0"		/* 5 */
		"ColLevel:\0"		/* 6 */
		"ColProg:\0"		/* 7 */
		"ColTitle:\0"		/* 8 */
		"ColTime:\0"		/* 9 */
		"ImageOffset:\0"	/* 10 */
		"RectTrack:\0"		/* 11 */
		"ColTrack:\0"		/* 12 */
		"ColButton:\0"		/* 13 */
		"Gain:\0"			/* 14 */
		"\0";
	static const WORD def[] = {	/* Default display layout */
		BTN_WIDTH, 0, LCD_WIDTH-1, LCD_HEIGHT-1, GRAY,
		50, 192, 309, 205, BLACK,
		50, 208, 150, 230, BLACK,
		155, 220, 309, 230, BLACK,
		155, 208, 229, 218, BLACK,
		233, 208, 309, 218, BLACK,
		0
	};

	/* Initialize configuration parameters in default setting */
	RectTitle.l = 51; RectTitle.t = 193; RectTitle.r = 308; RectTitle.b = 204;
	ColTitle[0] = BLACK; ColTitle[1] = I_WHITE; TitleAlign = 1;
	RectTrack.l = 51; RectTrack.t = 211; RectTrack.r = 80; RectTrack.b = 222;
	ColTrack[0] = BLACK; ColTrack[1] = I_CYAN;
	RectTime.l = 89; RectTime.t = 211; RectTime.r = 146; RectTime.b = 227;
	ColTime[0] = BLACK; ColTime[1] = I_GREEN; SmallTime = 0;
	RectProg.l = 156; RectProg.t = 222; RectProg.r = 308; RectProg.b = 228;
	ColProg[0] = GRAY; ColProg[1] = YELLOW;
	RectLevel[0].l = 156; RectLevel[0].t = 209; RectLevel[0].r = 229; RectLevel[0].b = 217; DirLevel[0] = 3;
	RectLevel[1].l = 233; RectLevel[1].t = 209; RectLevel[1].r = 308; RectLevel[1].b = 217; DirLevel[1] = 1;
	ColLevel[0] = BLACK; ColLevel[1] = I_GREEN; ColLevel[2] = I_YELLOW; ColLevel[3] = I_RED;
	ColButton[0] = WHITE; ColButton[1] = BLACK;
	imgofs.x = BTN_WIDTH; imgofs.y = 0;
	imgfile[0] = 0;
	Gain = 0;

	/* Load parameters from the configuration file */
	if (f_open(&fil, cfgfile, FA_READ) == FR_OK) {
		while (f_gets(line, sizeof(line), &fil)) {
			for (i = 0, pc = kwd; *pc; i++) {
				ps = line;
				while (*pc && tolower(*pc) == tolower(*ps) && *pc != ':') {
					pc++; ps++;
				}
				if (*pc == ':') { ps++; break; }
				while (*pc++);
			}
			switch (i) {
			case 0 :
				replace_filename(imgfile, sizeof(imgfile), cfgfile, ps);
				break;
			case 1 :
				if (get_rect(&ps, &RectLevel[0]) && xatoi(&ps, &val)) DirLevel[0] = (BYTE)val;
				break;
			case 2 :
				if (get_rect(&ps, &RectLevel[1]) && xatoi(&ps, &val)) DirLevel[1] = (BYTE)val;
				break;
			case 3 :
				if (get_rect(&ps, &RectTitle) && xatoi(&ps, &val)) TitleAlign = (BYTE)val;
				break;
			case 4 :
				get_rect(&ps, &RectProg);
				break;
			case 5 :
				if (get_rect(&ps, &RectTime) && xatoi(&ps, &val)) SmallTime = (BYTE)val;
				break;
			case 6 :
				get_color(&ps, ColLevel, 4);
				break;
			case 7 :
				get_color(&ps, ColProg, 2);
				break;
			case 8 :
				get_color(&ps, ColTitle, 2);
				break;
			case 9 :
				get_color(&ps, ColTime, 2);
				break;
			case 10 :
				get_point(&ps, &imgofs);
				break;
			case 11 :
				get_rect(&ps, &RectTrack);
				break;
			case 12 :
				get_color(&ps, ColTrack, 2);
				break;
			case 13 :
				get_color(&ps, ColButton, 2);
				break;
			case 14 :
				xatoi(&ps, &val); Gain = (BYTE)val;
				break;
			}
		}
		f_close(&fil);
	}

	w = def;	/* Draw default background image */
	while (*w) {
		rect.l = *w++; rect.t = *w++; rect.r = *w++; rect.b = *w++;
		gd_boxfill(&rect, *w++);
	}

	Volume = (Attenuation > Gain) ? Attenuation - Gain : 0;
	for (i = 1024, n = 0; n < Attenuation; n++) i = i * 913 / 1024;
	eeprom_read(EE_LVGAIN, 1, &lvl);
	GuiLevelGain = (UINT)lvl * 10000 / i;

	/* Load background image from bmp file */
	if (f_open(&fil, imgfile, FA_READ) == FR_OK) {
		gd_load_bmpfile(&fil, imgofs);
		f_close(&fil);
	} else {
		imgofs.x = 156; imgofs.y = 114;
		gd_put_str(imgofs, (BYTE*)"No Image", 80, BLACK);
	}
}



/*------------------------------------------------------------------*/
/* Put title and artist to the LCD                                  */
/*------------------------------------------------------------------*/

static
void put_title (
	const char* title,
	const char* artist
) {
	UINT n, w, x;
	char ttl[51], *d;
	POINT pnt;


	if (ColTitle[0] == 0) return;
	gd_boxfill(&RectTitle, ColTitle[0]);
	d = ttl; n = sizeof(ttl) - 1;
	while (*title && n) { *d++ = *title++; n--; }
	if (*artist) {
		title = " - ";
		while (*title && n) { *d++ = *title++; n--; }
		while (*artist && n) { *d++ = *artist++; n--; }
	}
	*d = 0;

	w = RectTitle.r - RectTitle.l + 1;
	x = RectTitle.l;
	switch (TitleAlign) {
	case 1:
		if (w > strlen(ttl) * ANK_W)
			x += (w - strlen(ttl) * ANK_W) / 2;
		break;
	case 2:
		if (w > strlen(ttl) * ANK_W)
			x += w - strlen(ttl) * ANK_W;
		break;
	}
	pnt.x = x;
	pnt.y = RectTitle.t;
	gd_put_str(pnt, (BYTE*)ttl, w, ColTitle[1]);
}



/*------------------------------------------------------------------*/
/* Put time to the LCD                                              */
/*------------------------------------------------------------------*/

static
void put_time (
	INT pt			/* Play time in ms (-1:blanked) */
) {
	INT m;
	POINT pnt;


	if (ColTime[0] == 0) return;
	gd_boxfill(&RectTime, ColTime[0]);
	if (pt == -1) return;
	pnt.x = RectTime.l; pnt.y = RectTime.t;
	m = pt / 60; pt %= 60;
	if (SmallTime) {
		if (m >= 10) gd_put_ank(pnt, m / 10 + '0', ColTime[1]);
		pnt.x += 6;
		gd_put_ank(pnt, m % 10 + '0', ColTime[1]);
		pnt.x += 6;
		gd_put_ank(pnt, ':', ColTime[1]);
		pnt.x += 6;
		gd_put_ank(pnt, pt / 10 + '0', ColTime[1]);
		pnt.x += 6;
		gd_put_ank(pnt, pt % 10 + '0', ColTime[1]);
	} else {
		if (m >= 10) gd_put_7seg(pnt, m / 10, ColTime[1]);
		pnt.x += 14;
		gd_put_7seg(pnt, m % 10, ColTime[1]);
		pnt.x += 14;
		gd_put_7seg(pnt, 10, ColTime[1]);
		pnt.x += 5;
		gd_put_7seg(pnt, pt / 10, ColTime[1]);
		pnt.x += 14;
		gd_put_7seg(pnt, pt % 10, ColTime[1]);
	}
}



/*------------------------------------------------------------------*/
/* Put a progress bar to the LCD                                    */
/*------------------------------------------------------------------*/

static
void put_prog (
	UINT pp,		/* Current position */
	UINT total		/* Total length */
) {
	UINT lw, mw;
	RECT rect;


	if (ColProg[0] == 0) return;
	mw = RectProg.r - RectProg.l + 1;
	if (total >= 1024)
		lw = mw * (pp / 1024) / (total /1024);
	else
		lw = 0;
	rect.t = RectProg.t;
	rect.b = RectProg.b;
	rect.l = RectProg.l;
	if (lw) {
		rect.r = rect.l + lw - 1;
		gd_boxfill(&rect, ColProg[1]);
	}
	mw -= lw;
	if (mw) {
		rect.r = RectProg.r;
		rect.l += lw;
		gd_boxfill(&rect, ColProg[0]);
	}
}



/*------------------------------------------------------------------*/
/* Put track number to the LCD                                      */
/*------------------------------------------------------------------*/

static
void put_track (
	UINT ctrk,		/* Current track */
	UINT ntrk		/* Number of tracks */
) {
	UINT d;
	char trk[6];
	POINT pnt;


	if (ColTrack[0] == 0) return;
	gd_boxfill(&RectTrack, ColTrack[0]);
	d = GuiCurrTrk;
	trk[0] = '0' + d / 10;
	trk[1] = '0' + d % 10;
	trk[2] = '/';
	d = GuiNumTrks;
	trk[3] = '0' + d / 10;
	trk[4] = '0' + d % 10;
	trk[5] = 0;
	pnt.x = RectTrack.l;
	pnt.y = RectTrack.t;
	gd_put_str(pnt, (BYTE*)trk, RectTrack.r - RectTrack.l + 1, ColTrack[1]);
}



/*------------------------------------------------------------------*/
/* Put level bar in the LCD                                         */
/*------------------------------------------------------------------*/

static
void put_level (
	INT ch,		/* Channel: 0=left, 1=right*/
	WORD lv,	/* Current level */
	WORD peak	/* Peak level */
) {
	UINT n, pk, lv60, lv80, lv100;
	RECT rect;
	BYTE col;


	if (ColLevel[0] == 0) return;
	n = (DirLevel[ch] == 2) ? RectLevel[ch].b - RectLevel[ch].t + 1: RectLevel[ch].r - RectLevel[ch].l + 1;
	lv60 = n * 6 / 10;
	lv80 = n * 8 / 10;
	lv100 = n;
	n = lv100 * lv / 65536;
	pk = lv100 * peak / 65536;

	switch (DirLevel[ch]) {
	case 1 :
		rect.t = RectLevel[ch].t;
		rect.b = RectLevel[ch].b;
		rect.l = RectLevel[ch].l;
		rect.r = (n >= lv60) ? rect.l + lv60 - 1 : rect.l + n - 1;
		gd_boxfill(&rect, ColLevel[1]);
		if (n > lv60) {
			rect.r = (n >= lv80) ? RectLevel[ch].l + lv80 - 1 : RectLevel[ch].l + n - 1;
			rect.l += lv60;
			gd_boxfill(&rect, ColLevel[2]);
			if (n > lv80) {
				rect.r = RectLevel[ch].l + n - 1;
				rect.l += lv80 - lv60;
				gd_boxfill(&rect, ColLevel[3]);
			}
		}
		rect.l = RectLevel[ch].l + n;
		rect.r = RectLevel[ch].r;
		gd_boxfill(&rect, ColLevel[0]);
		rect.r = RectLevel[ch].l + pk;
		rect.l = rect.r - 1;
		if (pk >= lv80) {
			col = ColLevel[3];
		} else if (pk >= lv60) {
			col = ColLevel[2];
		} else {
			col = ColLevel[1];
		}
		if (peak) gd_boxfill(&rect, col);
		break;
	case 2 :
		rect.l = RectLevel[ch].l;
		rect.r = RectLevel[ch].r;
		rect.b = RectLevel[ch].b;
		rect.t = (n >= lv60) ? rect.b - lv60 + 1 : rect.b - n + 1;
		gd_boxfill(&rect, ColLevel[1]);
		if (n > lv60) {
			rect.t = (n >= lv80) ? RectLevel[ch].b - lv80 + 1 : RectLevel[ch].b - n + 1;
			rect.b -= lv60;
			gd_boxfill(&rect, ColLevel[2]);
			if (n > lv80) {
				rect.t = RectLevel[ch].b - n + 1;
				rect.b -= lv80 - lv60;
				gd_boxfill(&rect, ColLevel[3]);
			}
		}
		rect.b = RectLevel[ch].b - n;
		rect.t = RectLevel[ch].t;
		gd_boxfill(&rect, ColLevel[0]);
		rect.t = RectLevel[ch].b - pk;
		rect.b = rect.t + 1;
		if (pk >= lv80) {
			col = ColLevel[3];
		} else if (pk >= lv60) {
			col = ColLevel[2];
		} else {
			col = ColLevel[1];
		}
		if (peak) gd_boxfill(&rect, col);
		break;
	case 3 :
		rect.t = RectLevel[ch].t;
		rect.b = RectLevel[ch].b;
		rect.r = RectLevel[ch].r;
		rect.l = (n >= lv60) ? rect.r - lv60 + 1 : rect.r - n + 1;
		gd_boxfill(&rect, ColLevel[1]);
		if (n > lv60) {
			rect.l = (n >= lv80) ? RectLevel[ch].r - lv80 + 1 : RectLevel[ch].r - n + 1;
			rect.r -= lv60;
			gd_boxfill(&rect, ColLevel[2]);
			if (n > lv80) {
				rect.l = RectLevel[ch].r - n + 1;
				rect.r -= lv80 - lv60;
				gd_boxfill(&rect, ColLevel[3]);
			}
		}
		rect.r = RectLevel[ch].r - n;
		rect.l = RectLevel[ch].l;
		gd_boxfill(&rect, ColLevel[0]);
		rect.l = RectLevel[ch].r - pk;
		rect.r = rect.l + 1;
		if (pk >= lv80) {
			col = ColLevel[3];
		} else if (pk >= lv60) {
			col = ColLevel[2];
		} else {
			col = ColLevel[1];
		}
		if (peak) gd_boxfill(&rect, col);
		break;
	}
}



/*------------------------------------------------------------------*/
/* Put a button to the LCD                                          */
/*------------------------------------------------------------------*/

static
void put_button (
	const char* mnu,	/* Menu strings */
	UINT item,			/* Item number */
	BYTE state			/* Button state (0:normal, 1:pressed) */
) {
	UINT n, f;
	RECT rect, r;
	POINT pnt;
	char s[10], c;


	rect.l = 0; rect.r = 39;
	rect.t = item * 40; rect.b = item * 40 + 39;
	gd_boxfill(&rect, ColButton[0]);
	if (state == 0) {
		r.l = rect.l; r.t = r.b = rect.t; r.r = rect.r;
		gd_boxfill(&r, I_WHITE);
		r.r = r.l; r.b = rect.b;
		gd_boxfill(&r, I_WHITE);
		r.t++; r.r = r.l = rect.r;
		gd_boxfill(&r, BLACK);
		r.t = r.b; r.l = rect.l + 1;
		gd_boxfill(&r, BLACK);
	} else {
		r.l = rect.l; r.t = r.b = rect.b; r.r = rect.r;
		gd_boxfill(&r, I_WHITE);
		r.l = r.r; r.t = rect.t;
		gd_boxfill(&r, I_WHITE);
		r.b--; r.r = r.l = rect.l;
		gd_boxfill(&r, BLACK);
		r.b = r.t; r.r = rect.r - 1;
		gd_boxfill(&r, BLACK);
	}

	while (item--) {
		do c = *mnu++; while ((BYTE)c >= ' ');
	}

	n = f = 0;
	do {
		c = *mnu++;
		if (c == '!') {
			f = 1; break;
		}
		if ((BYTE)c < ' ') break;
		s[n++] = c;
	} while (n < sizeof(s) - 1);
	s[n] = 0;
	if (n * ANK_W > rect.r - rect.l - 3) return;
	pnt.x = rect.l + (rect.r - rect.l + 1 - n * ANK_W) / 2;
	if (state) pnt.x++;
	if (!f) {
		pnt.y = rect.t + (rect.b - rect.t + 1 - ANK_H) / 2;
		if (state) pnt.y++;
	} else {
		pnt.y = rect.t + (rect.b - rect.t + 1 - ANK_H * 2) / 2;
		if (state) pnt.y++;
		gd_put_str(pnt, (BYTE*)s, n * ANK_W, ColButton[1]);
		n = 0;
		do {
			c = *mnu++;
			if (c == '!') {
				f = 1; break;
			}
			if ((BYTE)c < ' ') break;
			s[n++] = c;
		} while (n < sizeof(s) - 1);
		s[n] = 0;
		if (n * ANK_W > rect.r - rect.l - 3) return;
		pnt.x = rect.l + (rect.r - rect.l + 1 - n * ANK_W) / 2;
		if (state) pnt.x++;
		pnt.y = rect.t + (rect.b - rect.t + 1 - ANK_H * 2) / 2 + ANK_H;
		if (state) pnt.y++;
	}
	gd_put_str(pnt, (BYTE*)s, n * ANK_W, ColButton[1]);
}



/*------------------------------------------------------------------*/
/* Adjust LCD contrast                                              */
/*------------------------------------------------------------------*/

static
void adjust_cont (
	WORD ad		/* 10bit AD value (0..1023) */
)
{
	static const SHORT r_tbl[] =	/* 32 segments boundary element table (103JT with 10 kohms) */
	{ -490, -490, -358, -273, -207, -152, -105, -62,
	  -22, 15, 51, 85, 118, 151, 184, 217,
	  250, 284, 318, 354, 392, 431, 473, 519,
	  569, 624, 688, 763, 856, 976, 976, 976 };
	INT tmp, i;
	BYTE d;


	ad = ~ad;		/* PU/PD conversion (Not required when thermister is wired to high side */
	ad &= 1023;		/* Mask out invalid bits */
	if( (ad < 64) || (ad > 1024-64) ) return;				/* Abort when a sensor error, short or open, occured */
	i = ad >> 5; ad &= 31;									/* Select segment */
	tmp = (r_tbl[i] + (r_tbl[i+1] - r_tbl[i]) * ad / 32);	/* Get position in the segment */

	// 96/25.0/0deg 170/23.6/25deg 217/22.7/50deg
	if (tmp < 0) {
		d = 96;
	} else if (tmp < 250) {
		d = 96 + (170 - 96) * tmp / (250 - 0);
	} else if (tmp < 500) {
		d = 170 + (217 - 170) * (tmp - 250) / (500 - 250);
	} else {
		d = 217;
	}
	DA0CS0 = d + GuiContrast;
}


/*------------------------------------------------------------------*/
/* The task of display control                                      */
/*------------------------------------------------------------------*/

void TASK display (void* arg)
{
	char cfgfile[30];
	INT disptime, t, ch;
	UINT n;


	GuiCmd = 0;
	for (;;) {
		switch (GuiCmd) {
		case GI_INIT:	/* On wake up from stop state */
			gd_enable();
			CfgFile[0] = 0;
			break;

		case GI_NEW :	/* On start a new track */
			PlayState = 0;
			replace_filename(cfgfile, sizeof(cfgfile), (const char*)GuiFile, CNF_FILE);
			if (strcmp(CfgFile, cfgfile)) {
				strcpy(CfgFile, cfgfile);
				load_cfgfile(CfgFile);
			}
			adjust_cont(GuiLevel[2] >> 6);
			put_title((const char*)GuiTitle, (const char*)GuiArtist);
			PlayTime = disptime = 0;
			put_time(-1);
			put_track(GuiCurrTrk, GuiNumTrks);
			put_level(0, 0, 0);
			put_level(1, 0, 0);
			put_prog(0, 0);
			TabCmd = 1; wakeup_task(3);	/* Change button strings */
			break;

		case GI_PLAY :	/* On enter to play state */
			PlayState = 1;
			LevelLv[0] = LevelLv[1] = 0;
			LevelPk[0] = LevelPk[1] = 0;
			put_time(disptime);
			adjust_cont(GuiLevel[2] >> 6);
			break;

		case GI_STOP :	/* On enter to stop state */
			PlayState = 0;
			put_time(-1);
			put_level(0, 0, 0);
			put_level(1, 0, 0);
			put_prog(0, 0);
			TabCmd = 2; wakeup_task(3);	/* Change button strings */
			GuiCmd = 0;
			sleep(10000);
			if (GuiCmd) continue;
			gd_disable();
			break;

		case GI_PAUSE :	/* On enter to pause state */
			PlayState = 2;
			t = 0;
			break;

		default:	/* 20Hz interval timer */
			if (PlayState) {	/* In active state */
				if (PlayState == 1) {	/* In playback */
					t = PlayTime / 1000;
					if (t > disptime) {
						disptime = t;
						put_time(disptime);
						put_prog(GuiTrkPos, GuiTrkSize);
					}
				} else {				/* In pause */
					if ((t & 15) == 0)
						put_time((t & 16) ? disptime : -1);
					t++;
				}
				for (ch = 0; ch < 2; ch++) {
					if (LevelPkTmr[ch] == 0) {
						LevelPk[ch] = 0;
					} else {
						LevelPkTmr[ch]--;
					}
					if (LevelLv[ch] > LevelPk[ch]) {
						LevelPk[ch] = LevelLv[ch];
						LevelPkTmr[ch] = 20;
					}
					LevelLv[ch] = (LevelLv[ch] > LEVEL_DR) ? (LevelLv[ch] - LEVEL_DR) : 0;
					n = GuiLevel[ch] * GuiLevelGain / 1024;
					if (n > 65535) n = 65535;
					if (n > LevelLv[ch]) LevelLv[ch] = n;
					put_level(ch, LevelLv[ch], LevelPk[ch]);
				}
			}
		}
		/* Enter sleep state (waked up by new command or interval timer) */
		GuiCmd = 0;
		sleep(-1);
	}

}


/*------------------------------------------------------------------*/
/* The task of tablet control                                       */
/*------------------------------------------------------------------*/

void TASK tablet (void* arg)
{
	BOOL press;
	POINT pnt;
	UINT cmd;
	UINT n;
	static const char *mnu;


	for (;;) {
		sleep(-1);
		if (TabCmd) {
			sleep(50);
			mnu = (TabCmd == 2) ? menustr2 : menustr1;
			for (n = 0; n < 6; n++) put_button(mnu, n, 0);
			TabCmd = 0;
			pnt.x = 1000;
			continue;
		}
		if (GuiPoint.x < 1000 && !press) {
			pnt = GuiPoint;
			if (pnt.x <= 50) {
				cmd = pnt.y / 40;
				put_button(mnu, cmd, 1);
			}
			press = TRUE;
		} else if (GuiPoint.x == 1000 && press) {
			if (!gd_check()) {
				CmdButton = K_PLAY;
			} else if (pnt.x <= 50) {
				put_button(mnu, cmd, 0);
				CmdButton = cmd + 1;
			} else {
				CmdButton = PlayState ? K_STOP : K_PLAY;
			}
			press = FALSE;
		}
	}

}




