/*--------------------------------------------------------------------------*/
/*  IIC control for STA013 and 24Cxx                                        */

#include "iic.h"


#define SCL_LOW()	PM3H.1 = 0		/* SCL = LOW */
#define SCL_HIGH()	PM3H.1 = 1		/* SCL = High-Z */
#define	SCL_VAL		P3H.1			/* SCL input level */
#define SDA_LOW()	PM3H.0 = 0		/* SDA = LOW */
#define SDA_HIGH()	PM3H.0 = 1		/* SDA = High-Z */
#define	SDA_VAL		P3H.0			/* SDA input level */



static
void iic_delay (void)
{
	INT n, m;

	for (n = 3; n; n--) m = P3H;
}



static
void iic_start (void)
{
	SDA_HIGH();
	iic_delay();
	SCL_HIGH();
	iic_delay();
	SDA_LOW();
	iic_delay();
	SCL_LOW();
	iic_delay();
}



static
void iic_stop (void)
{
	SDA_LOW();
	iic_delay();
	SCL_HIGH();
	iic_delay();
	SDA_HIGH();
	iic_delay();
}



static
BOOL iic_send (BYTE dat)
{
	BYTE b = 0x80;
	BOOL ack;


	do {
		if (dat & b)	 {	/* SDA = Z/L */
			SDA_HIGH();
		} else {
			SDA_LOW();
		}
		iic_delay();
		SCL_HIGH();
		iic_delay();
		SCL_LOW();
		iic_delay();
	} while (b >>= 1);
	SDA_HIGH();
	iic_delay();
	SCL_HIGH();
	ack = SDA_VAL ? FALSE : TRUE;	/* Sample ACK */
	iic_delay();
	SCL_LOW();
	iic_delay();
	return ack;
}



static
BYTE iic_rcvr (BOOL ack)
{
	UINT d = 1;


	do {
		d <<= 1;
		SCL_HIGH();
		if (SDA_VAL) d++;
		iic_delay();
		SCL_LOW();
		iic_delay();
	} while (d < 0x100);
	if (ack) {		/* SDA = ACK */
		SDA_LOW();
	} else {
		SDA_HIGH();
	}
	iic_delay();
	SCL_HIGH();
	iic_delay();
	SCL_LOW();
	SDA_HIGH();
	iic_delay();

	return (BYTE)d;
}



void sta013_setregs (
	const BYTE* ptr		/* Setup data */
)
{
	BYTE reg, dat;
	BOOL ack;


	for (;;) {
		reg = *ptr++;
		dat = *ptr++;
		if (reg == 0xFF) break;
		do {
			iic_start();
			ack = iic_send(0x86);		/* STA013 address */
			if (ack) {
				ack = iic_send(reg);
				if (ack) ack = iic_send(dat);
			}
			iic_stop();
		} while (!ack);
	}
}



BOOL eeprom_read (
	UINT adr,		/* Read start address */
	UINT cnt,		/* Read byte count */
	void* buff		/* Read data buffer */
)
{
	BYTE *rbuff = buff;


	if (!cnt) return FALSE;				/* When no data to read, return */

	do {								/* Select EEPROM (0xA0) */
		iic_start();
	} while (!iic_send(0xA0));
	if (iic_send((BYTE)adr)) {			/* Set start address */
		iic_start();					/* Reselect EEPROM in read mode (0xA1) */
		if (iic_send(0xA1)) {
			do {						/* Receive data */
				cnt--;
				*rbuff++ = iic_rcvr(cnt ? TRUE : FALSE);
			} while (cnt);
		}
	}
	iic_stop();

	return cnt ? FALSE : TRUE;
}



BOOL eeprom_write (
	UINT adr,			/* Write start address */
	UINT cnt,			/* Write byte count */
	const void* buff	/* Data to be written */
)
{
	const BYTE *wbuff = buff;


	if (!cnt) return FALSE;				/* When no data to read, return */

	do {								/* Select EEPROM (0xA0) */
		iic_start();
	} while (!iic_send(0xA0));
	if (iic_send((BYTE)adr)) {			/* Set start address */
		do {							/* Send data */
			if (!iic_send(*wbuff++)) break;
		} while (--cnt);
	}
	iic_stop();

	return cnt ? FALSE : TRUE;
}



