/*---------------------------------------------------------------*/
/* Graphic MP3 Player                                            */
/*---------------------------------------------------------------*/


#include <string.h>
#include "comm.h"
#include "diskio.h"
#include "tff.h"
#include "taskmgr.h"
#include "glcd.h"
#include "gui.h"


int _rcopy(unsigned long*, long);
extern unsigned long _S_romp;


FATFS fatfs;				/* File system object */

volatile UINT Timer;		/* 1kHz increment timer */



/* Task stacks */
#define SS0	256		/* 1KB for player() */
#define SS1	256		/* 1KB for monitor() */
#define SS2	256		/* 1KB for display() */
#define SS3	256		/* 1KB for tablet() */

DWORD Stk0[SS0];
DWORD Stk1[SS1];
DWORD Stk2[SS2];
DWORD Stk3[SS3];


/* Task declarations */
void player (void*);	/* Task 0 */
void monitor (void*);	/* Task 1 */
void display (void*);	/* Task 2 */
void tablet (void*);	/* Task 3 */

DWORD get_fattime (void)
{
	return 0;
}



/*-----------------------------------------------------------------------*/
/* Main                                                                  */


int main ()
{
	PRCMD = 0; OCDM = 0;	/* Disable Debugging I/F */
	RCM = 1; WDTM2 = 0;		/* Disable WDT */

	VSWC = 0;				/* Set no wait state for peripherals */
	do; while(LOCKR);		/* Select clock source: PLLx4 */
	PLLCTL = 3;
	PRCMD = 0; PCC = 0;

	_rcopy(&_S_romp, -1);	/* Initialize .sdata sections */


	/* Initialize GPIO ports */
	P0  = 0x40;		/* P0: -HLZLL-- */
	PM0 = 0x93;
	PM1 = 0xFD;		/* P1: ------LZ */
	P3  = 0x0001;	/* P3: ------ZZ LLLLZLZH */
	PM3 = 0xFF0A;
	P4  = 0x02;		/* P4: -----LHZ */
	PM4 = 0xF9;
	P5  = 0x00;		/* P5: --LLLLLL */
	PM5 = 0xC0;
	PM7L = 0x1F;	/* P7L: LLLZZZZZ */
	P7H =  0x03;	/* P7H: ----LLHH */
	PM7H = 0xF0;
	PCM =  0x04;	/* PCM: ----LHLZ */
	PMCM = 0xF1;
	PCT = 0x13;		/* PCT: -L-H--HH */
	PMCT = 0xAC;
	PM9 = 0x0000;	/* P9: LLLLLLLL LLLLLLLL */
	PMDL = 0x0000;	/* PDL: LLLLLLLL LLLLLLLL */
	PMDH = 0xC0;	/* PDH: --LLLLLL */

	/* Enable external memory I/F */
	BSC = 0x5555;	/* Set bus width 16 bit */
	EXIMC = 0x01;	/* Separated bus mode */
	DWC0 = 0x7771;	/* Set bus wait state */
	AWC = 0xFFFE;
	BCC = 0xAAA8;
	PMC9 = 0xFFFF;	/* Attach address bus to I/O pad */
	PMCDH = 0x3F;
	PMCDL = 0xFFFF;	/* Attach data bus to I/O pad */
	PMCCM = 0x03;	/* Attach clkout,wait# to I/O pad */
	PMCCT = 0x13;	/* Attach conttol bus to I/O pad */
	PCM.2 = 0;		/* Enable LCDC CS# */

	/* Initialize MMC I/F */
	PMC4 = 0x07;			/* Attach CSIB0 unit to I/O pad */
	CB0CTL1 = 0x18;			/* Select fclk=fxx/2, SPI-0 */
	CB0CTL2 = 0;			/* Select 8 bit mode */
	CB0CTL0 = 0xE0;			/* Enable CSIB0 with single R/W, MSB first */

	/* Initialize decoder data I/F */
	PMC5 |= 0x30;			/* Attach CSIB2 unit to I/O pad */
	CB2CTL1 = 0x18;			/* Select fclk=fxx/2, SPI-0 */
	CB2CTL2 = 0;			/* Select 8 bit mode */
	CB2CTL0 = 0xE0;			/* Enable CSIB2 with single R/W, MSB first */
	PMC0.4 = 1;				/* Enable INT1 function with rising edge */
	INTR0.4 = 1;

	/* Initialize ADC */
	ADA0M0 = 0x30;	/* 1 shot scan mode */
	ADA0M1 = 0x03;	/* Conversion time */
	ADA0S = 4;		/* Scan range: ch0-ch4 */

	/* Initialize DAC */
	DA0CE0 = 1;

	/* Start TM0 in interval time of 1ms */
	TM0CMP0 = SYSCLK / 1000 - 1;
	TM0CTL0 = 0x80;
	TM0EQIC0 = 0x06;

	/* Initialize UARTA0 */
	PMC3L |= 0x03;			/* Attach UARTA0 unit to I/O pad */
	UA0CTL0 = 0x92;			/* Select serial format of N81,115.2kbps */
	UA0CTL1 = 0;
	UA0CTL2 = SYSCLK / 2 / 115200;
	UA0CTL0 = 0xF2;
	uart_init();

	__EI();


	/* Start threads */
	start_task(0, player,  &Stk0[SS0], NULL);
	start_task(1, monitor, &Stk1[SS1], NULL);
	start_task(2, display, &Stk2[SS2], NULL);
	start_task(3, tablet, &Stk3[SS3], NULL);
	for (;;) {
		dispatch_task(0);
		dispatch_task(1);
		dispatch_task(2);
		dispatch_task(3);
	}
}



/*---------------------------------------------------------*/
/* 1000Hz timer interrupt generated by Timer 0             */

#pragma interrupt INTTM0EQ0 ISR_tmm0
__multi_interrupt
void ISR_tmm0 (void)
{
	UINT n;
	static WORD level[4];
	static UINT tpph, tx, px, py;
	static BYTE np, chc;


	__EI();

	Timer++;			/* Performance counter for debug monitor */
	disk_timerproc();	/* Timer procedure of low level disk I/O module */
	task_timerproc();	/* Timer procedure of task manager */

	GuiLevel[2] = ADA0CR0;
	n = ADA0CR1;
	if (n > level[0]) level[0] = n;
	if (n < level[1]) level[1] = n;
	n = ADA0CR2;
	if (n > level[2]) level[2] = n;
	if (n < level[3]) level[3] = n;
	n = PlayState;
	if (n) {
		if (n == 1) PlayTime++;
		if (++TmrGui >= 20) {
			TmrGui = 0;
			GuiLevel[0] = level[0] - level[1];
			level[0] = 0; level[1] = 65535;
			GuiLevel[1] = level[2] - level[3];
			level[2] = 0; level[3] = 65535;
			wakeup_task(2);
		}
	}

	tpph++;
	switch (tpph & 3) {
	case 0:
		P7H = 0x09;
		n = ~(ADA0CR4 >> 6) & 1023;
		if (n < 30 || tx > 1010) np = 1;
		px += tx;
		n = (n > 47) ? n - 47 : 0;
		py += n;
		if ((tpph & 0x1C) == 0x1C) {
			if (!np) {
				if (chc) chc--;
			} else {
				if (chc < 2) chc++;
			}
			if (chc != 1) {
				GuiPoint.x = np ? 1000 : px * 320 / (1024 * 8);
				GuiPoint.y = py * 240 / (913 * 8);
				wakeup_task(3);
			}
			px = py = np = 0;
		}
		break;
	case 2:
		P7H = 0x06;
		tx = ADA0CR3 >> 6;
		break;
	}


	ADA0CE =1;

	__DI();
}

